from typing import TYPE_CHECKING
import h5py
import zarr

from .LindiH5pyDataset import LindiH5pyDataset
from .LindiH5pyLink import LindiH5pyHardLink, LindiH5pySoftLink
from .LindiH5pyAttributes import LindiH5pyAttributes


if TYPE_CHECKING:
    from .LindiH5pyFile import LindiH5pyFile  # pragma: no cover


class LindiH5pyGroup(h5py.Group):
    def __init__(self, _zarr_group: zarr.Group, _file: "LindiH5pyFile"):
        self._zarr_group = _zarr_group
        self._file = _file
        self._readonly = _file.mode not in ['r+']

        # In h5py, the id property is an object that exposes low-level
        # operations specific to the HDF5 library. LINDI aims to override the
        # high-level methods such that the low-level operations on id are not
        # needed. However, sometimes packages (e.g., pynwb) use the id as a
        # unique identifier for purposes of caching. Therefore, we make the id
        # to be a string that is unique for each object. If any of the low-level
        # operations are attempted on this id string, then an exception will be
        # raised, which will usually indicate that one of the high-level methods
        # should be overridden.
        self._id = f'{id(self._file)}/{self._zarr_group.name}'

        # The self._write object handles all the writing operations
        from .writers.LindiH5pyGroupWriter import LindiH5pyGroupWriter  # avoid circular import
        if self._readonly:
            self._writer = None
        else:
            self._writer = LindiH5pyGroupWriter(self)

    def __getitem__(self, name):
        if isinstance(name, (bytes, str)):
            x = self._zarr_group[name]
        else:
            raise TypeError(
                "Accessing a group is done with bytes or str, "
                "not {}".format(type(name))
            )
        if isinstance(x, zarr.Group):
            # follow the link if this is a soft link
            soft_link = x.attrs.get('_SOFT_LINK', None)
            if soft_link is not None:
                link_path = soft_link['path']
                target_item = self._file.get(link_path)
                if not isinstance(target_item, (LindiH5pyGroup, LindiH5pyDataset)):
                    raise Exception(
                        f"Expected a group or dataset at {link_path} but got {type(target_item)}"
                    )
                return target_item
            return LindiH5pyGroup(x, self._file)
        elif isinstance(x, zarr.Array):
            return LindiH5pyDataset(x, self._file)
        else:
            raise Exception(f"Unknown type: {type(x)}")

    def get(self, name, default=None, getclass=False, getlink=False):
        if not (getclass or getlink):
            try:
                return self[name]
            except KeyError:
                return default

        if name not in self:
            return default
        elif getclass and not getlink:
            raise Exception("Getting class is not allowed")
        elif getlink and not getclass:
            x = self._zarr_group.get(name, default=default)
            if x is None:
                return default
            soft_link = x.attrs.get('_SOFT_LINK', None)
            if isinstance(x, zarr.Group) and soft_link is not None:
                return LindiH5pySoftLink(soft_link['path'])
            else:
                return LindiH5pyHardLink()
        else:
            raise Exception("Impossible")

    @property
    def name(self):
        return self._zarr_group.name

    def keys(self):  # type: ignore
        return self._zarr_group.keys()

    def __iter__(self):
        return self._zarr_group.__iter__()

    def __reversed__(self):
        raise Exception("Not implemented: __reversed__")

    def __contains__(self, name):
        return self._zarr_group.__contains__(name)

    def __str__(self):
        return f'<{self.__class__.__name__}: {self.name}>'

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.name}>'

    @property
    def id(self):
        # see comment above
        return self._id

    @property
    def file(self):
        return self._file

    @property
    def attrs(self):  # type: ignore
        return LindiH5pyAttributes(self._zarr_group.attrs, readonly=self._file.mode == 'r')

    @property
    def ref(self):
        if self._readonly:
            raise ValueError("Cannot get ref on read-only object")
        assert self._writer is not None
        return self._writer.ref

    ##############################
    # write
    def create_group(self, name, track_order=None):
        if self._readonly:
            raise Exception('Cannot create group in read-only mode')
        assert self._writer is not None
        return self._writer.create_group(name, track_order=track_order)

    def require_group(self, name):
        if self._readonly:
            raise Exception('Cannot require group in read-only mode')
        assert self._writer is not None
        return self._writer.require_group(name)

    def create_dataset(self, name, shape=None, dtype=None, data=None, **kwds):
        if self._readonly:
            raise Exception('Cannot create dataset in read-only mode')
        assert self._writer is not None
        return self._writer.create_dataset(name, shape=shape, dtype=dtype, data=data, **kwds)

    def require_dataset(self, name, shape, dtype, exact=False, **kwds):
        if self._readonly:
            raise Exception('Cannot require dataset in read-only mode')
        assert self._writer is not None
        return self._writer.require_dataset(name, shape, dtype, exact=exact, **kwds)

    def __setitem__(self, name, obj):
        if self._readonly:
            raise Exception('Cannot set item in read-only mode')
        assert self._writer is not None
        return self._writer.__setitem__(name, obj)

    def __delitem__(self, name):
        if self._readonly:
            raise Exception('Cannot delete item in read-only mode')
        assert self._writer is not None
        return self._writer.__delitem__(name)



from typing import Union
from dataclasses import dataclass


@dataclass(frozen=True)
class LindiH5ZarrStoreOpts:
    """
    Options for the LindiH5ZarrStore class.

    Attributes:
        num_dataset_chunks_threshold (Union[int, None]): For each dataset in the
        HDF5 file, if the number of chunks is greater than this threshold, then
        the dataset will be represented as an external array link. If None, then
        no datasets will be represented as external array links (equivalent to a
        threshold of 0). Default is 1000.
    """
    num_dataset_chunks_threshold: Union[int, None] = 1000



from typing import Union
import os


class LocalCache:
    def __init__(self, *, cache_dir: Union[str, None] = None):
        if cache_dir is None:
            # use ~/.lindi/cache as default cache directory
            cache_dir = os.path.expanduser("~/.lindi/cache")
        self._cache_dir = cache_dir
        os.makedirs(self._cache_dir, exist_ok=True)
        self._sqlite_db_fname = os.path.join(self._cache_dir, "lindi_cache.db")
        self._sqlite_client = LocalCacheSQLiteClient(db_fname=self._sqlite_db_fname)

    def get_remote_chunk(self, *, url: str, offset: int, size: int):
        return self._sqlite_client.get_remote_chunk(url=url, offset=offset, size=size)

    def put_remote_chunk(self, *, url: str, offset: int, size: int, data: bytes):
        if len(data) != size:
            raise ValueError("data size does not match size")
        self._sqlite_client.put_remote_chunk(url=url, offset=offset, size=size, data=data)


class ChunkTooLargeError(Exception):
    pass


class LocalCacheSQLiteClient:
    def __init__(self, *, db_fname: str):
        import sqlite3
        self._db_fname = db_fname
        self._conn = sqlite3.connect(self._db_fname)
        self._cursor = self._conn.cursor()
        self._cursor.execute(
            """
            PRAGMA journal_mode=WAL
            """
        )
        self._cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS remote_chunks (
                url TEXT,
                offset INTEGER,
                size INTEGER,
                data BLOB,
                PRIMARY KEY (url, offset, size)
            )
            """
        )
        self._conn.commit()

    def get_remote_chunk(self, *, url: str, offset: int, size: int):
        self._cursor.execute(
            """
            SELECT data FROM remote_chunks WHERE url = ? AND offset = ? AND size = ?
            """,
            (url, offset, size),
        )
        row = self._cursor.fetchone()
        if row is None:
            return None
        return row[0]

    def put_remote_chunk(self, *, url: str, offset: int, size: int, data: bytes):
        if size >= 1000 * 1000 * 900:
            # This is a sqlite limitation/configuration
            # https://www.sqlite.org/limits.html
            # For some reason 1000 * 1000 * 1000 seems to be too large, whereas 1000 * 1000 * 900 is fine
            raise ChunkTooLargeError("Cannot store blobs larger than 900 MB in LocalCache")
        self._cursor.execute(
            """
            INSERT OR REPLACE INTO remote_chunks (url, offset, size, data) VALUES (?, ?, ?, ?)
            """,
            (url, offset, size, data),
        )
        self._conn.commit()



import numpy as np


def decode_nan_inf_ninf(val):
    if isinstance(val, list):
        return [decode_nan_inf_ninf(v) for v in val]
    elif isinstance(val, dict):
        return {k: decode_nan_inf_ninf(v) for k, v in val.items()}
    elif val == 'NaN':
        return float('nan')
    elif val == 'Infinity':
        return float('inf')
    elif val == '-Infinity':
        return float('-inf')
    else:
        return val


def encode_nan_inf_ninf(val):
    if isinstance(val, list):
        return [encode_nan_inf_ninf(v) for v in val]
    elif isinstance(val, dict):
        return {k: encode_nan_inf_ninf(v) for k, v in val.items()}
    elif isinstance(val, (float, np.floating)):
        if np.isnan(val):
            return 'NaN'
        elif val == float('inf'):
            return 'Infinity'
        elif val == float('-inf'):
            return '-Infinity'
        else:
            return val
    else:
        return val



from typing import Any, TYPE_CHECKING
import h5py
import zarr
import numpy as np

from ..LindiH5pyReference import LindiH5pyReference
from ...conversion._util import _is_numeric_dtype
from ...conversion.create_zarr_dataset_from_h5_data import h5_object_data_to_zarr_data

if TYPE_CHECKING:
    from ..LindiH5pyDataset import LindiH5pyDataset  # pragma: no cover


class LindiH5pyDatasetWriter:
    def __init__(self, p: 'LindiH5pyDataset'):
        self.p = p

    def __setitem__(self, args, val):
        if isinstance(self.p._zarr_array, h5py.Dataset):
            self.p._zarr_array.__setitem__(args, val)
        elif isinstance(self.p._zarr_array, zarr.Array):
            self._set_item_for_zarr(self.p._zarr_array, args, val)
        else:
            raise Exception(f"Unexpected type: {type(self.p._zarr_array)}")

    def _set_item_for_zarr(self, zarr_array: zarr.Array, selection: Any, val: Any):
        if self.p._compound_dtype is not None:
            raise Exception("Setting compound dataset is not implemented")
        if self.p.ndim == 0:
            if selection != ():
                raise TypeError(f'Cannot slice a scalar dataset with {selection}')
            zarr_array[0] = val
        else:
            dtype = zarr_array.dtype
            if _is_numeric_dtype(dtype) or dtype in [bool, np.bool_]:
                # this is the usual numeric case
                zarr_array[selection] = val
            elif dtype.kind == 'O':
                zarr_array[selection] = h5_object_data_to_zarr_data(val, h5f=None, label='')
            else:
                raise Exception(f'Unsupported dtype for slice setting {dtype} in {self.p.name}')

    @property
    def ref(self):
        return LindiH5pyReference({
            'object_id': self.p.attrs.get('object_id', None),
            'path': self.p.name,
            'source': '.',
            'source_object_id': self.p.file.attrs.get('object_id', None)
        })



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..LindiH5pyAttributes import LindiH5pyAttributes  # pragma: no cover


class LindiH5pyAttributesWriter:
    def __init__(self, p: 'LindiH5pyAttributes'):
        self.p = p

    def __setitem__(self, key, value):
        from ...conversion.attr_conversion import h5_to_zarr_attr  # avoid circular import
        if self.p._readonly:
            raise KeyError("Cannot set attributes on read-only object")
        self.p._attrs[key] = h5_to_zarr_attr(value, h5f=None)



from typing import TYPE_CHECKING
import h5py
import numpy as np
import zarr
import numcodecs
from numcodecs.abc import Codec

from ..LindiH5pyDataset import LindiH5pyDataset
from ..LindiH5pyReference import LindiH5pyReference

if TYPE_CHECKING:
    from ..LindiH5pyGroup import LindiH5pyGroup  # pragma: no cover

from ...conversion.create_zarr_dataset_from_h5_data import create_zarr_dataset_from_h5_data

_compression_not_specified_ = object()


class LindiH5pyGroupWriter:
    def __init__(self, p: 'LindiH5pyGroup'):
        self.p = p

    def create_group(self, name, track_order=None):
        from ..LindiH5pyGroup import LindiH5pyGroup  # avoid circular import
        if track_order is not None:
            raise Exception("track_order is not supported (I don't know what it is)")
        if isinstance(self.p._zarr_group, h5py.Group):
            return LindiH5pyGroup(
                self.p._zarr_group.create_group(name), self.p._file
            )
        elif isinstance(self.p._zarr_group, zarr.Group):
            return LindiH5pyGroup(
                self.p._zarr_group.create_group(name), self.p._file
            )
        else:
            raise Exception(f'Unexpected group object type: {type(self.p._zarr_group)}')

    def require_group(self, name):
        if name in self.p:
            ret = self.p[name]
            if not isinstance(ret, LindiH5pyGroup):
                raise Exception(f'Expected a group at {name} but got {type(ret)}')
            return ret
        return self.create_group(name)

    def create_dataset(
        self,
        name,
        shape=None,
        dtype=None,
        data=None,
        **kwds
    ):
        chunks = None
        compression = _compression_not_specified_
        compression_opts = None
        for k, v in kwds.items():
            if k == 'chunks':
                chunks = v
            elif k == 'compression':
                compression = v
            elif k == 'compression_opts':
                compression_opts = v
            else:
                raise Exception(f'Unsupported kwds in create_dataset: {k}')

        if compression is _compression_not_specified_:
            _zarr_compressor = 'default'
            if compression_opts is not None:
                raise Exception('compression_opts is only supported when compression is provided')
        elif isinstance(compression, Codec):
            _zarr_compressor = compression
            if compression_opts is not None:
                raise Exception('compression_opts is not supported when compression is provided as a Codec')
        elif isinstance(compression, str):
            if compression == 'gzip':
                if compression_opts is None:
                    level = 4  # default for h5py
                elif isinstance(compression_opts, int):
                    level = compression_opts
                else:
                    raise Exception(f'Unexpected type for compression_opts: {type(compression_opts)}')
                _zarr_compressor = numcodecs.GZip(level=level)
            else:
                raise Exception(f'Compression {compression} is not supported')
        else:
            raise Exception(f'Unexpected type for compression: {type(compression)}')

        if isinstance(self.p._zarr_group, h5py.Group):
            if _zarr_compressor != 'default':
                raise Exception('zarr_compressor is not supported when _group_object is h5py.Group')
            return LindiH5pyDataset(
                self._group_object.create_dataset(name, shape=shape, dtype=dtype, data=data, chunks=chunks),  # type: ignore
                self.p._file
            )
        elif isinstance(self.p._zarr_group, zarr.Group):
            if isinstance(data, list):
                data = np.array(data)
            if shape is None:
                if data is None:
                    raise Exception('shape or data must be provided')
                if isinstance(data, np.ndarray):
                    shape = data.shape
                else:
                    shape = ()
            if dtype is None:
                if data is None:
                    raise Exception('dtype or data must be provided')
                if isinstance(data, np.ndarray):
                    dtype = data.dtype
                else:
                    dtype = np.dtype(type(data))
            ds = create_zarr_dataset_from_h5_data(
                zarr_parent_group=self.p._zarr_group,
                name=name,
                label=(self.p.name or '') + '/' + name,
                h5_chunks=chunks,
                h5_shape=shape,
                h5_dtype=dtype,
                h5_data=data,
                h5f=None,
                zarr_compressor=_zarr_compressor
            )
            return LindiH5pyDataset(ds, self.p._file)
        else:
            raise Exception(f'Unexpected group object type: {type(self.p._zarr_group)}')

    def require_dataset(self, name, shape, dtype, exact=False, **kwds):
        if name in self.p:
            ret = self.p[name]
            if not isinstance(ret, LindiH5pyDataset):
                raise Exception(f'Expected a dataset at {name} but got {type(ret)}')
            if ret.shape != shape:
                raise Exception(f'Expected shape {shape} but got {ret.shape}')
            if exact:
                if ret.dtype != dtype:
                    raise Exception(f'Expected dtype {dtype} but got {ret.dtype}')
            else:
                if not np.can_cast(ret.dtype, dtype):
                    raise Exception(f'Cannot cast dtype {ret.dtype} to {dtype}')
            return ret
        return self.create_dataset(name, *(shape, dtype), **kwds)

    def __setitem__(self, name, obj):
        if isinstance(obj, h5py.SoftLink):
            if isinstance(self.p._zarr_group, h5py.Group):
                self.p._zarr_group[name] = obj
            elif isinstance(self.p._zarr_group, zarr.Group):
                grp = self.p.create_group(name)
                grp._zarr_group.attrs['_SOFT_LINK'] = {
                    'path': obj.path
                }
            else:
                raise Exception(f'Unexpected group object type: {type(self.p._zarr_group)}')
        else:
            raise Exception(f'Unexpected type for obj in __setitem__: {type(obj)}')

    def __delitem__(self, name):
        del self.p._zarr_group[name]

    @property
    def ref(self):
        return LindiH5pyReference({
            'object_id': self.p.attrs.get('object_id', None),
            'path': self.p.name,
            'source': '.',
            'source_object_id': self.p.file.attrs.get('object_id', None)
        })



from typing import Union
import json


def reformat_json(x: Union[bytes, None]) -> Union[bytes, None]:
    """Reformat to not include whitespace and to not allow nan, inf, and ninf.

    It is assumed that float attributes nan, inf, and ninf float values have
    been encoded as strings. See encode_nan_inf_ninf() and h5_to_zarr_attr().
    """
    if x is None:
        return None
    a = json.loads(x.decode("utf-8"))
    return json.dumps(a, separators=(",", ":"), allow_nan=False).encode("utf-8")
